/*=========================================================================
 *
 *  Copyright NumFOCUS
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/

#include "PerformanceBenchmarkingInformation.h"

#include "PerformanceBenchmarkingUtilities.h"
#include "PerformanceBenchmarkingInformation.h"

#include "itkObjectFactory.h"

#include <ostream>
#include <fstream>
#include <sstream>
#include <algorithm>


// Construct const versions via the emplace method.  This macro simplifies the text
// written in ITK/Modules/Core/Common/src/CMakeLists.txt.
#define MAKE_MAP_ENTRY(KEY, VALUE, DESCRIPTION )                                          \
m_Map.emplace(                                                                            \
  std::pair< MapKeyType, InformationValueType>(                                           \
    MapKeyType{ KEY },                                                                    \
    InformationValueType{ MapValueType{ VALUE }, MapValueDescriptionType{ DESCRIPTION } } \
  )                                                                                       \
)

namespace itk {

std::mutex PerformanceBenchmarkingInformation::m_Mutex;
PerformanceBenchmarkingInformation::Pointer PerformanceBenchmarkingInformation::m_InformationInstance;
PerformanceBenchmarkingInformation::MapType PerformanceBenchmarkingInformation::m_Map;

PerformanceBenchmarkingInformation::Pointer
PerformanceBenchmarkingInformation
::New()
{
  return Self::GetInstance();
}

PerformanceBenchmarkingInformation::Pointer
PerformanceBenchmarkingInformation
::GetInstance()
{
  std::lock_guard<std::mutex> mutexHolder(m_Mutex);
  if (m_InformationInstance.IsNull())
  {
    m_InformationInstance = ObjectFactory<Self>::Create();
    {
      new PerformanceBenchmarkingInformation(); //constructor sets m_InformationInstance
    }
  }
  return m_InformationInstance;
}

const PerformanceBenchmarkingInformation::MapType &
PerformanceBenchmarkingInformation::GetMap()
{
  return PerformanceBenchmarkingInformation::GetInstance()->m_Map;
}

const PerformanceBenchmarkingInformation::MapValueType
PerformanceBenchmarkingInformation::GetValue(const MapKeyType &&key)
{
  const MapType &localMap = PerformanceBenchmarkingInformation::GetInstance()->GetMap();

  auto it = localMap.find(key);
  if (it != localMap.end())
  {
    return it->second.m_Value;
  }
  return {""};
}

const PerformanceBenchmarkingInformation::MapValueDescriptionType
PerformanceBenchmarkingInformation::GetDescription(const MapKeyType && key)
{
  const MapType &localMap = PerformanceBenchmarkingInformation::GetInstance()->GetMap();

  auto it = localMap.find(key);
  if (it != localMap.end())
  {
    return it->second.m_Description;
  }
  return {""};
}

const std::vector<  PerformanceBenchmarkingInformation::MapKeyType >
PerformanceBenchmarkingInformation::GetAllKeys()
{
  std::vector< PerformanceBenchmarkingInformation::MapKeyType > keyVector;
  keyVector.reserve(30);
  for( auto elem : PerformanceBenchmarkingInformation::GetInstance()->m_Map )
  {
    keyVector.emplace_back( elem.first );
  }
  return keyVector;
}

PerformanceBenchmarkingInformation
::PerformanceBenchmarkingInformation()
{
  m_InformationInstance = this; //threads need this
  m_InformationInstance->UnRegister(); // Remove extra reference

  
  MAKE_MAP_ENTRY( "CONFIGURE_DATE",        "2021-07-30 07:29",       "The date of configuration."              );
  MAKE_MAP_ENTRY( "PROJECT_URL",           "https://github.com/InsightSoftwareConsortium/ITKPerformanceBenchmarking",
                                                                                      "The URL of project."              );

  MAKE_MAP_ENTRY( "GIT_CONFIG_SHA1",       "GIT-NOTFOUNDGIT-NOTFOUND",
                                                                               "The git HEAD SHA1 has at configuration." );
  MAKE_MAP_ENTRY( "GIT_CONFIG_REFSPEC",    "GIT-NOTFOUND",   "The git HEAD ref spec."                  );
  MAKE_MAP_ENTRY( "GIT_CONFIG_DATE",       "GIT-NOTFOUND",      "The git HEAD date"                       );
  MAKE_MAP_ENTRY( "GIT_VERSION_DESC",      "GIT-NOTFOUND",     "The git HEAD version tag-commits-hash"   );

  MAKE_MAP_ENTRY( "CMAKE_VERSION",         "3.20.5",        "The version of cmake used to configure." );
  MAKE_MAP_ENTRY( "CMAKE_GENERATOR",       "Unix Makefiles",      "The generator used for building."        );
  MAKE_MAP_ENTRY( "CMAKE_BUILD_TYPE",      "RelWithDebInfo",     "The build type requested."               );

  MAKE_MAP_ENTRY( "CMAKE_C_COMPILER_ID",   "GNU",  "The C compiler ID used for building."    );
  MAKE_MAP_ENTRY( "CMAKE_C_COMPILER_VERSION",   "10.3.1",
                                                                          "The C compiler version used for building."    );
  MAKE_MAP_ENTRY( "CMAKE_C_COMPILER",      "/usr/bin/cc",     "The C compiler."                       );
  MAKE_MAP_ENTRY( "CMAKE_C_FLAGS",         "-pipe -frecord-gcc-switches -Wall -g -O2 -fomit-frame-pointer -march=armv7-a -mthumb  -Wno-uninitialized -Wno-unused-parameter -Wcast-align -Wdisabled-optimization -Wextra -Wformat=2 -Winvalid-pch -Wno-format-nonliteral -Wpointer-arith -Wshadow -Wunused -Wwrite-strings -funit-at-a-time -Wno-strict-overflow  ",
                                                                          "The C compiler flags."                 );

  MAKE_MAP_ENTRY( "CMAKE_CXX_STANDARD",    "11",   "The CXX standard specified."             );
  MAKE_MAP_ENTRY( "CMAKE_CXX_COMPILER_ID", "GNU","The CXX compiler ID used for building."  );
  MAKE_MAP_ENTRY( "CMAKE_CXX_COMPILER_VERSION",   "10.3.1",
                                                                          "The CXX compiler version used for building."  );
  MAKE_MAP_ENTRY( "CMAKE_CXX_COMPILER",    "/usr/bin/c++",   "The CXX compiler."                       );
  MAKE_MAP_ENTRY( "CMAKE_CXX_FLAGS",       "-std=gnu++11 -pipe -frecord-gcc-switches -Wall -g -O2 -fomit-frame-pointer -march=armv7-a -mthumb  -Wcast-align -Wdisabled-optimization -Wextra -Wformat=2 -Winvalid-pch -Wno-format-nonliteral -Wpointer-arith -Wshadow -Wunused -Wwrite-strings -funit-at-a-time -Wno-strict-overflow -Wno-deprecated -Wno-invalid-offsetof -Woverloaded-virtual -Wstrict-null-sentinel  ",
                                                                          "The CXX compiler flags."                 );

}

} //end namespace itk
